import { Observable, pipe, Subject, BehaviorSubject, queueScheduler, isObservable, EMPTY, identity, from, of } from 'rxjs';
import { filter, map, distinctUntilChanged, catchError, observeOn, withLatestFrom, tap, finalize, mergeMap } from 'rxjs/operators';

class StoreExtension {
  constructor() {
    this.sortOrder = 0
    /* DEFAULT */
    ;
  }

}
class Actions extends Observable {}
const EFFECT_METADATA_KEY = '@mini-rx/effectMetaData';

const miniRxNameSpace = '@mini-rx';

function createMiniRxActionType(miniRxActionType, featureKey) {
  return miniRxNameSpace + (featureKey ? '/' + featureKey : '') + '/' + miniRxActionType;
}

function createSetStateAction(stateOrCallback, featureId, featureKey, name) {
  const miniRxActionType = "set-state"
  /* SET_STATE */
  ;
  return {
    type: createMiniRxActionType(miniRxActionType, featureKey) + (name ? '/' + name : ''),
    stateOrCallback,
    featureId,
    featureKey,
    miniRxActionType
  };
}
function createMiniRxAction(miniRxActionType, featureKey) {
  return {
    type: createMiniRxActionType(miniRxActionType, featureKey)
  };
}
const key = 'miniRxActionType';
const type = "set-state"
/* SET_STATE */
; // Type predicate

function isSetStateAction(action) {
  return action[key] === type;
}

function ofType(...allowedTypes) {
  return filter(action => allowedTypes.some(type => {
    return type === action.type;
  }));
}
function select(mapFn) {
  return pipe(map(mapFn), distinctUntilChanged());
}
function miniRxError(message) {
  throw new Error(miniRxNameSpace + ': ' + message);
}
function miniRxConsoleError(message, err) {
  console.error(miniRxNameSpace + ': ' + message + '\nDetails:', err);
}
/** @internal */

function hasEffectMetaData(param) {
  return param.hasOwnProperty(EFFECT_METADATA_KEY);
} // Simple alpha numeric ID: https://stackoverflow.com/a/12502559/453959
// This isn't a real GUID!

function generateId() {
  return Math.random().toString(36).slice(2);
}
function beautifyActionForLogging(action, state) {
  if (isSetStateAction(action)) {
    return mapSetStateActionToActionWithPayload(action, state);
  }

  return action;
}

function mapSetStateActionToActionWithPayload(action, state) {
  const stateOrCallback = action.stateOrCallback;
  const featureState = state[action.featureKey];
  return {
    type: action.type,
    payload: typeof stateOrCallback === 'function' ? stateOrCallback(featureState) : stateOrCallback
  };
}

// Credits go to NgRx

function defaultEffectsErrorHandler(observable$, retryAttemptLeft = 10) {
  return observable$.pipe(catchError(error => {
    miniRxConsoleError(`An error occurred in the Effect. MiniRx resubscribed the Effect automatically and will do so ${retryAttemptLeft - 1} more times.\nPlease provide error handling inside the Effect using \`catchError\` or \`tapResponse\`.`, error);

    if (retryAttemptLeft <= 1) {
      return observable$; // last attempt
    } // Return observable that produces this particular effect


    return defaultEffectsErrorHandler(observable$, retryAttemptLeft - 1);
  }));
}

// Credits go to NgRx
// Copied from with small modifications: https://github.com/ngrx/platform/blob/12.0.0/modules/store/src/utils.ts
function combineReducers(reducers) {
  const reducerKeys = Object.keys(reducers);
  const reducerKeyLength = reducerKeys.length;
  return (state = {}, action) => {
    const stateKeysLength = Object.keys(state).length;
    let hasChanged = stateKeysLength !== reducerKeyLength;
    const nextState = {};

    for (let i = 0; i < reducerKeyLength; i++) {
      const key = reducerKeys[i];
      const reducer = reducers[key];
      const previousStateForKey = state[key];
      const nextStateForKey = reducer(previousStateForKey, action);
      nextState[key] = nextStateForKey;
      hasChanged = hasChanged || nextStateForKey !== previousStateForKey;
    }

    return hasChanged ? nextState : state;
  };
}

class StoreCore {
  constructor() {
    // ACTIONS
    this.actionsSource = new Subject();
    this.actions$ = this.actionsSource.asObservable(); // APP STATE

    this.stateSource = new BehaviorSubject({}); // Init App State with empty object

    this.state$ = this.stateSource.asObservable(); // REDUCERS (Feature state reducers and meta reducers)

    this.reducerStateSource = new BehaviorSubject({
      featureReducers: {},
      metaReducers: []
    }); // EXTENSIONS

    this.extensions = [];
    let reducer; // 👇 Refactored `withLatestFrom` in actions$.pipe to own subscription (fewer operators = less bundle-size :))

    this.reducerStateSource.subscribe(v => {
      const combinedMetaReducer = combineMetaReducers(v.metaReducers);
      const combinedReducer = combineReducers(v.featureReducers);
      reducer = combinedMetaReducer(combinedReducer);
    }); // Listen to the Actions Stream and update state accordingly

    this.actions$.pipe(observeOn(queueScheduler) // Prevent stack overflow: https://blog.cloudboost.io/so-how-does-rx-js-queuescheduler-actually-work-188c1b46526e
    ).subscribe(action => {
      const newState = reducer(this.stateSource.getValue(), action);
      this.updateState(newState);
    });
  }

  get reducerState() {
    return this.reducerStateSource.getValue();
  }

  get featureReducers() {
    return this.reducerState.featureReducers;
  }

  set featureReducers(featureReducers) {
    this.reducerStateSource.next(Object.assign(Object.assign({}, this.reducerState), {
      featureReducers
    }));
  }

  config(config = {}) {
    var _a, _b;

    if (Object.keys(this.featureReducers).length) {
      miniRxError('`configureStore` detected reducers. Did you instantiate FeatureStores before calling `configureStore`?');
    }

    if ((_a = config.metaReducers) === null || _a === void 0 ? void 0 : _a.length) {
      this.addMetaReducers(...config.metaReducers);
    }

    if ((_b = config.extensions) === null || _b === void 0 ? void 0 : _b.length) {
      const sortedExtensions = sortExtensions(config.extensions);
      sortedExtensions.forEach(extension => this.addExtension(extension));
    }

    if (config.reducers) {
      Object.keys(config.reducers).forEach(featureKey => {
        checkFeatureExists(featureKey, this.featureReducers);
        this.addReducer(featureKey, config.reducers[featureKey]); // config.reducers! (prevent TS2532: Object is possibly 'undefined')
      });
    }

    if (config.initialState) {
      this.updateState(config.initialState);
    }

    this.dispatch(createMiniRxAction("init-store"
    /* INIT_STORE */
    ));
  }

  addFeature(featureKey, reducer, config = {}) {
    var _a;

    reducer = ((_a = config.metaReducers) === null || _a === void 0 ? void 0 : _a.length) ? combineMetaReducers(config.metaReducers)(reducer) : reducer;

    if (!config.multi) {
      checkFeatureExists(featureKey, this.featureReducers);
    } else {
      featureKey = featureKey + '-' + generateId();
    }

    if (typeof config.initialState !== 'undefined') {
      reducer = createReducerWithInitialState(reducer, config.initialState);
    }

    this.addReducer(featureKey, reducer);
    this.dispatch(createMiniRxAction("init-feature"
    /* INIT_FEATURE */
    , featureKey));
    return featureKey;
  }

  removeFeature(featureKey) {
    this.removeReducer(featureKey);
    this.dispatch(createMiniRxAction("destroy-feature"
    /* DESTROY_FEATURE */
    , featureKey));
  }

  dispatch(action) {
    this.actionsSource.next(action);
  }

  updateState(state) {
    this.stateSource.next(state);
  }

  select(mapFn) {
    return this.state$.pipe(select(mapFn));
  }

  effect(effect$) {
    const effectWithErrorHandler$ = defaultEffectsErrorHandler(effect$);
    effectWithErrorHandler$.subscribe(action => {
      let dispatch = true;

      if (hasEffectMetaData(effect$)) {
        const metaData = effect$[EFFECT_METADATA_KEY];
        dispatch = !!metaData.dispatch;
      }

      if (dispatch) {
        this.dispatch(action);
      }
    });
  }

  addMetaReducers(...reducers) {
    this.reducerStateSource.next(Object.assign(Object.assign({}, this.reducerState), {
      metaReducers: [...this.reducerState.metaReducers, ...reducers]
    }));
  }

  addExtension(extension) {
    extension.init();
    this.extensions.push(extension);
  }

  addReducer(featureKey, reducer) {
    this.featureReducers = Object.assign(Object.assign({}, this.featureReducers), {
      [featureKey]: reducer
    });
  }

  removeReducer(featureKey) {
    this.featureReducers = omit(this.featureReducers, featureKey);
  }

}

function createReducerWithInitialState(reducer, initialState) {
  return (state = initialState, action) => {
    return reducer(state, action);
  };
}

function checkFeatureExists(featureKey, reducers) {
  if (reducers.hasOwnProperty(featureKey)) {
    miniRxError(`Feature "${featureKey}" already exists.`);
  }
}

function sortExtensions(extensions) {
  return [...extensions].sort((a, b) => {
    return a.sortOrder - b.sortOrder;
  });
}

function omit(object, keyToOmit) {
  return Object.keys(object).filter(key => key !== keyToOmit).reduce((prevValue, key) => {
    prevValue[key] = object[key];
    return prevValue;
  }, {});
}

function combineMetaReducers(metaReducers) {
  return reducer => {
    return metaReducers.reduceRight((previousValue, currentValue) => {
      return currentValue(previousValue);
    }, reducer);
  };
} // Created once to initialize singleton

/** @internal */


var StoreCore$1 = new StoreCore();

class Store {
  // Prevent direct construction calls with the `new` operator.
  constructor(config) {
    this.dispatch = StoreCore$1.dispatch.bind(StoreCore$1);
    this.select = StoreCore$1.select.bind(StoreCore$1);
    this.effect = StoreCore$1.effect.bind(StoreCore$1);
    StoreCore$1.config(config);
  } // Public Store API


  feature(featureKey, reducer, config) {
    StoreCore$1.addFeature(featureKey, reducer, config);
  }
  /** @internal */

  /** @deprecated This is an internal implementation detail, do not use. */


  static configureStore(config) {
    if (!Store.instance) {
      Store.instance = new Store(config);
      return Store.instance;
    }

    miniRxError('`configureStore` was called multiple times.');
  }

}
Store.instance = undefined;
function configureStore(config) {
  return Store.configureStore(config);
}
const actions$ = StoreCore$1.actions$;

// Credits go to Brecht Billiet
const defaultBufferSize = 100;
let isUndoExtensionInitialized;
let executedActions = [];
let initialState;
let bufferSize;
class UndoExtension extends StoreExtension {
  constructor(config = {
    bufferSize: defaultBufferSize
  }) {
    super();
    this.sortOrder = 1
    /* UNDO_EXTENSION */
    ;
    bufferSize = config.bufferSize;
  }

  init() {
    StoreCore$1.addMetaReducers(undoMetaReducer);
    isUndoExtensionInitialized = true;
  }

}
const UNDO_ACTION = miniRxNameSpace + '/undo';
function undo(action) {
  return {
    type: UNDO_ACTION,
    payload: action
  };
}

function undoMetaReducer(rootReducer) {
  return (state = {}, action) => {
    if (action.type === UNDO_ACTION) {
      // if the action is UNDO_ACTION,
      // then call all the actions again on the rootReducer,
      // except the one we want to rollback
      let newState = initialState;
      executedActions = executedActions.filter(eAct => eAct !== action.payload); // update the state for every action until we get the
      // exact same state as before, but without the action we want to rollback

      executedActions.forEach(executedAction => {
        newState = rootReducer(newState, executedAction);
      });
      return newState;
    } else {
      // push every action that isn't UNDO_ACTION
      executedActions.push(action);
    }

    const updatedState = rootReducer(state, action);

    if (executedActions.length === bufferSize + 1) {
      const firstAction = executedActions[0]; // calculate the state x (buffersize) actions ago

      initialState = rootReducer(initialState, firstAction); // keep the correct actions

      executedActions = executedActions.slice(1, bufferSize + 1);
    }

    return updatedState;
  };
}

class FeatureStore {
  constructor(featureKey, initialState, config = {}) {
    this.stateSource = new BehaviorSubject({});
    this.state$ = this.stateSource.asObservable();
    this.featureId = generateId();
    this._featureKey = StoreCore$1.addFeature(featureKey, createFeatureReducer(this.featureId, initialState), config);
    this.sub = StoreCore$1.select(state => state[this.featureKey]).subscribe(this.stateSource);
  }

  get state() {
    return this.stateSource.getValue();
  }

  get featureKey() {
    return this._featureKey;
  }

  setState(stateOrCallback, name) {
    const action = createSetStateAction(stateOrCallback, this.featureId, this.featureKey, name);
    StoreCore$1.dispatch(action);
    return action;
  }

  select(mapFn) {
    if (!mapFn) {
      return this.state$;
    }

    return this.state$.pipe(select(mapFn));
  }

  effect(effectFn) {
    const subject = new Subject();
    const effect$ = effectFn(subject);
    const effectWithDefaultErrorHandler = defaultEffectsErrorHandler(effect$);
    this.sub.add(effectWithDefaultErrorHandler.subscribe());
    return observableOrValue => {
      isObservable(observableOrValue) ? this.sub.add(observableOrValue.subscribe(v => subject.next(v))) : subject.next(observableOrValue);
    };
  }

  undo(action) {
    isUndoExtensionInitialized ? StoreCore$1.dispatch(undo(action)) : miniRxError('UndoExtension is not initialized.');
  }

  destroy() {
    this.sub.unsubscribe();
    StoreCore$1.removeFeature(this._featureKey);
  } // tslint:disable-next-line:use-lifecycle-interface


  ngOnDestroy() {
    this.destroy();
  }

}

function createFeatureReducer(featureId, initialState) {
  return (state = initialState, action) => {
    if (isSetStateAction(action) && action.featureId === featureId) {
      const stateOrCallback = action.stateOrCallback;
      const newPartialState = typeof stateOrCallback === 'function' ? stateOrCallback(state) : stateOrCallback;
      return Object.assign(Object.assign({}, state), newPartialState);
    }

    return state;
  };
}

function createFeatureStore(featureKey, initialState, config = {}) {
  return new FeatureStore(featureKey, initialState, config);
}

// Credits go to Alexander Reardon
// Copied from with small modifications: https://github.com/alexreardon/memoize-one/tree/v6.0.0/src
// MIT License
//
// Copyright (c) 2019 Alexander Reardon
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
function isEqual(first, second) {
  if (first === second) {
    return true;
  } // Special case for NaN (NaN !== NaN)


  if (Number.isNaN(first) && Number.isNaN(second)) {
    return true;
  }

  return false;
}

function areInputsEqual(newInputs, lastInputs) {
  // no checks needed if the inputs length has changed
  if (newInputs.length !== lastInputs.length) {
    return false;
  } // Using for loop for speed. It generally performs better than array.every
  // https://github.com/alexreardon/memoize-one/pull/59


  for (let i = 0; i < newInputs.length; i++) {
    if (!isEqual(newInputs[i], lastInputs[i])) {
      return false;
    }
  }

  return true;
}

function memoizeOne(resultFn) {
  let cache = null; // breaking cache when context (this) or arguments change

  function memoized(...newArgs) {
    if (cache && cache.lastThis === this && areInputsEqual(newArgs, cache.lastArgs)) {
      return cache.lastResult;
    } // Throwing during an assignment aborts the assignment: https://codepen.io/alexreardon/pen/RYKoaz
    // Doing the lastResult assignment first so that if it throws
    // the cache will not be overwritten


    const lastResult = resultFn.apply(this, newArgs);
    cache = {
      lastResult,
      lastArgs: newArgs,
      lastThis: this
    };
    return lastResult;
  }

  return memoized;
}

// Credits go to NgRx
function createSelector(...args) {
  const selectors = args.slice(0, args.length - 1);
  const projector = args[args.length - 1];
  const memoizedProjector = memoizeOne(projector);
  return memoizeOne(state => {
    const selectorResults = selectors.map(fn => fn(state));
    return memoizedProjector.apply(null, selectorResults);
  });
}
function createFeatureSelector(featureKey) {
  return createSelector(state => {
    if (featureKey) {
      return state[featureKey];
    }

    return state;
  }, featureState => featureState);
}

const defaultOptions = {
  name: 'MiniRx - Redux DevTools',
  traceLimit: 25
};
class ReduxDevtoolsExtension extends StoreExtension {
  constructor(options) {
    super();
    this.options = options;

    if (!window) {
      miniRxError('The Redux DevTools are only supported in browser environments.');
    }

    this.devtoolsExtension = window.__REDUX_DEVTOOLS_EXTENSION__;
    this.options = Object.assign(Object.assign({}, defaultOptions), this.options);
  }

  init() {
    if (this.devtoolsExtension) {
      this.devtoolsConnection = this.devtoolsExtension.connect(this.options);
      StoreCore$1.actions$.pipe(withLatestFrom(StoreCore$1.state$), tap(([action, state]) => {
        const actionForDevTools = beautifyActionForLogging(action, state);
        this.devtoolsConnection.send(actionForDevTools, state);
      })).subscribe();
      this.devtoolsConnection.subscribe(this.onDevToolsMessage.bind(this));
    }
  }

  onDevToolsMessage(message) {
    if (message.type === "DISPATCH"
    /* DISPATCH */
    ) {
      switch (message.payload.type) {
        case "JUMP_TO_STATE"
        /* JUMP_TO_STATE */
        :
        case "JUMP_TO_ACTION"
        /* JUMP_TO_ACTION */
        :
          this.updateState(JSON.parse(message.state));
      }
    }
  }

  updateState(state) {
    StoreCore$1.updateState(state);
  }

}

class LoggerExtension extends StoreExtension {
  init() {
    StoreCore$1.addMetaReducers(loggerMetaReducer);
  }

}

function loggerMetaReducer(reducer) {
  return (state, action) => {
    let actionToLog = beautifyActionForLogging(action, state);
    const nextState = reducer(state, action);
    console.log('%c' + action.type, 'color: #25c2a0', '\nAction:', actionToLog, '\nState: ', nextState);
    return nextState;
  };
}

// Copied from with small modifications: https://github.com/jsdf/deep-freeze/blob/v1.1.1/index.js
// Original License:
// This software is released to the public domain.
//
// It is based in part on the deepFreeze function from:
// https://developer.mozilla.org/en-US/docs/JavaScript/Reference/Global_Objects/Object/freeze
//
// https://developer.mozilla.org/en-US/docs/Project:Copyrights
function deepFreeze(o) {
  Object.freeze(o);
  const oIsFunction = typeof o === 'function';
  const hasOwnProp = Object.prototype.hasOwnProperty;
  Object.getOwnPropertyNames(o).forEach(function (prop) {
    if (hasOwnProp.call(o, prop) && (oIsFunction ? prop !== 'caller' && prop !== 'callee' && prop !== 'arguments' : true) && o[prop] !== null && (typeof o[prop] === 'object' || typeof o[prop] === 'function') && !Object.isFrozen(o[prop])) {
      deepFreeze(o[prop]);
    }
  });
  return o;
}

// Credits go to Brandon Roberts
class ImmutableStateExtension extends StoreExtension {
  init() {
    StoreCore$1.addMetaReducers(storeFreeze);
  }

}
function storeFreeze(reducer) {
  return (state = {}, action) => {
    deepFreeze(state);
    const nextState = reducer(state, action);
    deepFreeze(nextState);
    return nextState;
  };
}

// Credits go to NgRx
function tapResponse(fn1OrObject, fn2, fn3) {
  return source => {
    const tapResponseObj = typeof fn1OrObject === 'function' ? {
      next: fn1OrObject,
      error: fn2,
      finalize: fn3
    } : fn1OrObject;
    return source.pipe(tap({
      next: v => {
        if (tapResponseObj.next) {
          tryCatch(() => tapResponseObj.next(v), 'next');
        }
      },
      error: err => tryCatch(() => tapResponseObj.error(err), 'error')
    }), catchError(() => EMPTY), tapResponseObj.finalize ? finalize(() => {
      tryCatch(() => tapResponseObj.finalize(), 'finalize');
    }) : identity // Conditionally apply operator: https://rxjs.dev/api/index/function/identity
    );
  };
}

function tryCatch(fn, whichCallback) {
  try {
    fn();
  } catch (error) {
    miniRxConsoleError(`An error occurred in the \`tapResponse\` ${whichCallback} callback.`, error);
  }
}

function mapResponse(mapFn, errorFn) {
  return source => source.pipe(map(v => mapFn(v)), mergeMap(mapFnResult => {
    return Array.isArray(mapFnResult) ? from(mapFnResult) : of(mapFnResult);
  }), catchError(err => {
    const errorFnResult = errorFn(err);
    return errorFnResult ? Array.isArray(errorFnResult) ? from(errorFnResult) : of(errorFnResult) : EMPTY;
  }));
}

// Credits go to NgRx
const DEFAULT_EFFECT_CONFIG = {
  dispatch: true
};
function createEffect(v, config) {
  const value = Object.assign(Object.assign({}, DEFAULT_EFFECT_CONFIG), config);
  Object.defineProperty(v, EFFECT_METADATA_KEY, {
    value
  });
  return v;
}

export { Actions, FeatureStore, ImmutableStateExtension, LoggerExtension, ReduxDevtoolsExtension, Store, StoreExtension, UndoExtension, StoreCore$1 as _StoreCore, actions$, configureStore, createEffect, createFeatureSelector, createFeatureStore, createSelector, hasEffectMetaData, mapResponse, ofType, tapResponse, undo };
